"""
    Copyright Engie Impact Sustainability Solution EMEAI 2020.
    All rights reserved.
"""

__author__ = 'Engie Impact Sustainability Solution EMEAI'

import decimal
from unittest.mock import patch

import boto3
from boto3.dynamodb import types


class FloatSerializer(types.TypeSerializer):
    """
    This class can be used to patch the default dynamo serializer.
    Example:
        with patch('boto3.dynamodb.types.TypeSerializer', new=FloatSerializer), \
            patch('boto3.dynamodb.types.TypeDeserializer', new=FloatDeserializer):
            dynamodb = boto3.resource("dynamodb")
    """

    # Workaround to serialize float to Decimal
    # Credits: https://github.com/boto/boto3/issues/665#issuecomment-559892366
    def _is_number(self, value):
        if isinstance(value, (int, decimal.Decimal, float)):
            return True
        return False

    def _serialize_n(self, value):
        if isinstance(value, float):
            with decimal.localcontext(types.DYNAMODB_CONTEXT) as context:
                context.traps[decimal.Inexact] = 0
                context.traps[decimal.Rounded] = 0
                number = str(context.create_decimal_from_float(value))
                return number

        number = super(FloatSerializer, self)._serialize_n(value)
        return number

    def _serialize_m(self, value):
        return {str(k): self.serialize(v) for k, v in value.items()}


class FloatDeserializer(types.TypeDeserializer):
    def _deserialize_n(self, value):
        return float(value)


def create_resource_with_float_serializer():
    """
    Return a patched version of the boto3 dynamodb resource.
    This version of the resource will automatically convert float to Decimal and the reverse.

    :return: A DynamoDB resource who use float instead of Decimal.
    :rtype: DynamoDB
    """
    with patch("boto3.dynamodb.types.TypeSerializer", new=FloatSerializer), \
         patch("boto3.dynamodb.types.TypeDeserializer", new=FloatDeserializer):
        return boto3.resource("dynamodb")


def query_all_items(dynamo_resource, table_name, **kwargs):
    """
    Iterate over all the pages in the query result and concatenate them.

    :param dynamo_resource: The DynamoDB resource.
    :type dynamo_resource: DynamoDB
    :param table_name: the table name
    :type table_name: str
    :param kwargs: the query arguments. See Boto3 documentation.
    :return: All items in the table matching the query.
    :rtype: list
    """
    table = dynamo_resource.Table(table_name)
    response = table.query(**kwargs)
    items = response['Items']

    while response.get('LastEvaluatedKey'):
        kwargs['ExclusiveStartKey'] = response['LastEvaluatedKey']
        response = table.query(**kwargs)
        items.extend(response['Items'])

    return items



"""
    Copyright Engie Impact 2020.
    All rights reserved.
"""

__author__ = 'Engie Impact'

import logging

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import ChunkedEncodingError
from urllib3 import Retry

_logger = logging.getLogger(__name__)


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = 5
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


class SessionRetry(requests.Session):
    MAX_GLOBAL_RETRY = 3

    def request(self, method, url, params=None, data=None, headers=None, cookies=None, files=None, auth=None,
                timeout=None, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None,
                json=None):
        retry_attempt = 0

        while True:
            retry_attempt += 1
            try:
                return super().request(method, url, params, data, headers, cookies, files, auth, timeout,
                                       allow_redirects, proxies, hooks, stream, verify, cert, json)

            except ChunkedEncodingError as cee:
                if retry_attempt <= self.MAX_GLOBAL_RETRY and method != "POST":
                    _logger.debug(
                        f"Connection issue while performing request: retry ({retry_attempt})\ncause:{str(cee)}"
                    )
                    continue
                else:
                    raise cee

    def post(self, url, data=None, json=None, **kwargs):
        return super().post(url, data, json, **kwargs)


def create_requests_session(headers=None, api_gateway_key=None, max_retry=10, backoff_factor=1, timeout=5):
    """
    Create a requests session with custom retry policy.
    It will retry automatically if the http status code is 502, 503, 504.
    It will also retry on connection issues (except for post requests)
    You can also provide some headers to be used or/and an API key.

    :param headers: the headers that must be used.
    :type headers: dict | None
    :param api_gateway_key: An api key that will be set in the "x-api-key" header.
      If set this value override the value in the "headers" parameter.
    :type api_gateway_key: str | None
    :param max_retry: Total number of retries to allow.
    :type max_retry: int | None
    :param backoff_factor: A backoff factor to apply between attempts after the second try.
      urllib3 will sleep for:: {backoff factor} * (2 ** ({number of total retries} - 1)) seconds.
      It will never be longer than 2 minutes.
    :type backoff_factor: float | None
    :param timeout: How many seconds to wait for the server to send data before giving up
    :type timeout: int | None
    :return: the requests session
    :rtype: SessionRetry
    """
    session = SessionRetry()

    retry_policy = Retry(
        total=max_retry,
        backoff_factor=backoff_factor,  # exponential managed by the number of retry
        status_forcelist=[502, 503, 504]
    )
    session.mount('https://', TimeoutHTTPAdapter(max_retries=retry_policy, timeout=timeout))

    if headers:
        session.headers.update(headers)

    if api_gateway_key is not None:
        session.headers['x-api-key'] = api_gateway_key

    return session



"""
    Copyright Engie Impact Sustainability Solution EMEAI 2020.
    All rights reserved.
"""

__author__ = 'Engie Impact Sustainability Solution EMEAI'

import logging


class EIBError(Exception):
    """
    Application basic error. Should not have direct inheritance.
    Use instead BackendError or ClientError.
    """
    log_level = logging.ERROR
    title = None
    http_status = 500

    def __init__(self, detail):
        super().__init__(detail)
        self.detail = detail


class BackendError(EIBError):
    """
    Internal error (not suppose to be send outside of the service).
    """
    title = 'API Internal Error'


class ClientError(EIBError):
    log_level = logging.WARNING
    title = "API Error"
    http_status = 400

    def __init__(self, detail, http_status=None, title=None):
        """
        Client error that must be send to the end-users as API response.

        :param detail: A human-readable explanation specific to this occurrence of the problem.
        :type detail: str
        :param http_status: the http status code to use for this error.
          If not provided, the http_status define at the class level will be use.
          If not define on subclass then 400 will be used.
        :type http_status: int | None
        :param title: A short, human-readable summary of the problem type.
          It SHOULD NOT change from occurrence to occurrence of the problem.
          If not provided it will use the one define at the class level.
        :type title: str
        """
        super().__init__(detail)
        if http_status is not None:
            self.http_status = http_status

        if title is not None:
            self.title = title


class BadRequestError(ClientError):
    title = 'Bad Request'
    http_status = 400


class NotFoundError(ClientError):
    title = 'Not Found'
    http_status = 404



"""
    Copyright Engie Impact Sustainability Solution EMEAI 2020.
    All rights reserved.
"""

__author__ = 'Engie Impact Sustainability Solution EMEAI'

import json
import logging
from datetime import datetime
from decimal import Decimal
from functools import wraps

from eib_aws_utils.errors import EIBError, BackendError
from eib_aws_utils.logging import configure_logging

_logger = logging.getLogger(__name__)


def http_endpoint(main_module_name):

    def decorator(function):

        @wraps(function)
        def inner1(*args, **kwargs):
            context = kwargs.get('context') or (len(args) >= 2 and args[1]) or None
            configure_logging(main_module_name, context)

            try:
                response = function(*args, **kwargs)

                if not isinstance(response, tuple):
                    body, status = response, 200
                else:
                    body, status = response

                return _to_api_response(body, status)

            except EIBError as error:
                _logger.log(error.log_level, f"[{error.title} ({error.http_status})] {error.detail}")
                return _to_api_response(error, error.http_status)

            except Exception as error:
                _logger.exception(f"An unhandled exception occurs: {error}")
                returned_exception = BackendError("Unhandled Exception")
                return _to_api_response(returned_exception, 500)

        return inner1
    return decorator


def _to_api_response(body, status):
    return dict(
        statusCode=str(status),
        body=json.dumps(body, default=_custom_class_dumper),
        headers={
            'Content-Type': 'application/json',
            'Access-Control-Allow-Origin': '*'
        }
    )


def _custom_class_dumper(o):
    if isinstance(o, datetime):
        return o.isoformat()

    elif isinstance(o, Decimal):
        return float(o)

    elif isinstance(o, EIBError):
        return dict(
            title=o.title,
            status=o.http_status,
            detail=o.detail
        )

    try:
        return o.to_dict()
    except AttributeError:
        pass

    raise TypeError(f"Object of type '{o.__class__.__name__}' is not JSON serializable")



"""
    Copyright Engie Impact Sustainability Solution EMEAI 2020.
    All rights reserved.
"""

__author__ = 'Engie Impact Sustainability Solution EMEAI'

import logging
from functools import wraps

from eib_aws_utils.errors import EIBError
from eib_aws_utils.logging import configure_logging

_logger = logging.getLogger(__name__)


def entry_point(main_module_name):
    def decorator(function):

        @wraps(function)
        def inner1(*args, **kwargs):
            context = kwargs.get('context') or (len(args) >= 2 and args[1]) or None
            configure_logging(main_module_name, context)

            try:
                return function(*args, **kwargs)

            except EIBError as error:
                _logger.log(error.log_level, f"[{error.title} ({error.http_status})] {error.detail}")
                raise error

            except Exception as error:
                _logger.exception(f"An unhandled exception occurs: {error}")
                raise error

        return inner1
    return decorator



"""
    Copyright Engie Impact Sustainability Solution EMEAI 2020.
    All rights reserved.
"""

__author__ = 'Engie Impact Sustainability Solution EMEAI'

import setuptools

with open("README.md", "r") as fh:
    readme = fh.read()
with open("CHANGELOG.md", "r") as fh:
    changelog = fh.read()

long_description = "{}\n\n{}".format(readme, changelog)

setuptools.setup(
    name="eib-aws-utils",
    version="1.0.4",
    author=__author__,
    author_email="guido.maurano@engie.com",
    description="Helper used by Engie Impact Sustainability Solution EMEAI to develop services based on AWS",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.tools.digital.engie.com/TractebelHQImpulse/eib-aws-utils",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'python-json-logger',
        'requests'
    ]
)



"""
    Copyright Engie Impact Sustainability Solution EMEAI 2020.
    All rights reserved.
"""

__author__ = 'Engie Impact Sustainability Solution EMEAI'

import logging
import logging.config
import os
from uuid import uuid4

from pythonjsonlogger import jsonlogger


class CustomJsonFormatter(jsonlogger.JsonFormatter):
    context = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._context_id = None

    def _get_context_id(self):
        if self._context_id is None:
            if self.context and hasattr(self.context, 'aws_request_id'):
                self._context_id = self.context.aws_request_id
            else:
                self._context_id = str(uuid4())

        return self._context_id

    def add_fields(self, log_record, record, message_dict):
        super(CustomJsonFormatter, self).add_fields(log_record, record, message_dict)
        log_record['aws_request_id'] = self._get_context_id()


def configure_logging(main_module_name, context):
    """
    Configure the standard logging package to use JSON format.
    It also add the aws_request_id to the log fields.
    And use the LOGGING_LEVEL environment variable to define the logging level for the main_module_name package.

    :param main_module_name: the main package name
    :type main_module_name: str
    :param context: the aws context of the lambda
    :type context: object
    """
    logging.config.dictConfig(_get_default_logging_config(main_module_name))
    CustomJsonFormatter.context = context


def _get_default_logging_config(main_module_name):
    return {
        'version': 1,
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'default',
                'level': 'DEBUG'
            },
        },
        'formatters': {
            'default': {
                'format': '%(asctime)s %(name)s %(levelname)s %(lineno)s %(module)s %(message)s',
                'datefmt': '%Y-%m-%dT%H:%M:%S%z',
                'class': 'eib_aws_utils.logging.CustomJsonFormatter'
            }
        },
        'disable_existing_loggers': False,
        'root': {
            'level': 'INFO',
            'handlers': ['console']
        },
        'loggers': {
            main_module_name: {
                'level': os.getenv('LOGGING_LEVEL', 'INFO')
            }
        }
    }
